/**************************************************************************
** Filename: COM.cpp
**
** Title: Series 5000 Eyetracker Software
**
** Description:
** This file contains routines to read and write a serial port.
**
***************************************************************************
**
** Modification History:
**    1.0,  01/31/97, THC, Initial version.
**
**************************************************************************/

/* includes */

#include <vcl\vcl.h>
#pragma hdrstop
#include <stdio.h>

#include "COM.h"

/* constants */

#define INPUT_QUEUE_SIZE  512//1024
#define OUTPUT_QUEUE_SIZE 512//1024

/* functions */

__fastcall TCOM::TCOM(TComponent* Owner) : TComponent(Owner){
  FOpened = False;}


__fastcall TCOM::~TCOM()  {
  close();}

bool __fastcall TCOM::initialize( AnsiString aCOM,  int rate ){
  DCB          dcb;
  COMMTIMEOUTS CommTimeouts;
  if ( FOpened ) {
    close();
    FOpened = false;
  }
  FCOM = aCOM;
  FBaudRate = rate;
//  hCom = CreateFile( &FCOM[1], GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0,NULL );
  hCom = CreateFile( FCOM.c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0,NULL );
  if ( hCom == INVALID_HANDLE_VALUE ) return false;
  FOpened = True;
  SetupComm( hCom, INPUT_QUEUE_SIZE, OUTPUT_QUEUE_SIZE );
  GetCommState( hCom, &dcb );
  dcb.BaudRate = FBaudRate;
  dcb.ByteSize = 8;
  dcb.Parity = NOPARITY;
  dcb.StopBits = ONESTOPBIT;
  dcb.fBinary = TRUE;
  dcb.fOutxCtsFlow = FALSE;
  dcb.fOutxDsrFlow = FALSE;
  dcb.fDsrSensitivity = FALSE;
  dcb.fOutX = FALSE;
  dcb.fInX = FALSE;
  dcb.fParity = FALSE;
  dcb.fNull = FALSE;
  dcb.fRtsControl = RTS_CONTROL_DISABLE;
  dcb.fAbortOnError = FALSE;
  if( !SetCommState( hCom, &dcb))return false;
  GetCommTimeouts( hCom, &CommTimeouts );
  CommTimeouts . ReadIntervalTimeout = 1; //20;
  CommTimeouts . ReadTotalTimeoutMultiplier = 1;//20;
  CommTimeouts . ReadTotalTimeoutConstant = 0;
  CommTimeouts . WriteTotalTimeoutMultiplier = 0;
  CommTimeouts . WriteTotalTimeoutConstant = 1;//100;
  if( !SetCommTimeouts( hCom, &CommTimeouts ))return false;
  reset();
  return true ;
}


void __fastcall TCOM::close(){
	if ( FOpened ){
	   PurgeComm( hCom, PURGE_TXABORT	| PURGE_RXABORT );  // Flush queues
           CloseHandle( hCom );
        }
        FOpened=false;
}


void __fastcall TCOM::SetBaudRate( int rate ){
	DCB dcb;
	FBaudRate = rate;
	if ( !FOpened ) return;
	GetCommState( hCom, &dcb );
	dcb.BaudRate = FBaudRate;
	SetCommState( hCom, &dcb);}

void __fastcall TCOM::SetCOM( AnsiString aCOM ){
    initialize( aCOM, FBaudRate);}

void __fastcall TCOM::reset(){
	if ( FOpened  )
      PurgeComm( hCom, PURGE_TXABORT | PURGE_RXABORT ); }

void __fastcall TCOM::flush_input(){
	if ( FOpened ) PurgeComm( hCom, PURGE_RXABORT ); }

void __fastcall TCOM::flush_output(){
	if ( FOpened) PurgeComm( hCom, PURGE_TXABORT );}

int __fastcall TCOM::get_input_count(){
	DWORD   error_code;
	COMSTAT ComStat;
	if ( !FOpened )return 0;
	if( !ClearCommError( hCom, &error_code, &ComStat ))
      return 0;
	return ComStat . cbInQue;}

int __fastcall TCOM::get_output_space(){
	DWORD   error_code;
	COMSTAT ComStat;
	if ( !FOpened )	return 0;
 	ClearCommError( hCom, &error_code, &ComStat );
	return ( OUTPUT_QUEUE_SIZE - ComStat . cbOutQue );}

Byte __fastcall TCOM::read_byte(){
	DWORD   num;
    Byte b=0;
	if ( FOpened ) ReadFile( hCom, &b, 1, &num, NULL );
    return b;}

bool __fastcall TCOM::write_byte( const Byte& data ){
    DWORD num=0;
    if ( !FOpened ) return false;
    WriteFile( hCom, &data, 1, &num, NULL );
    return (num>0);
}

bool __fastcall TCOM::read_buffer( Byte *buffer, int count, DWORD *actualRead )
{
	if ( !FOpened ) return false;
    ReadFile( hCom, buffer, count, actualRead, NULL );
	return true;
}

bool __fastcall TCOM::write_buffer (Byte *buffer, int count, DWORD *actualWrite )
{
	if ( !FOpened )
	   	return false;

	int result = WriteFile( hCom, buffer, count, actualWrite, NULL );
	return true;

}
//**********************************

