//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <stdlib.h>
#include <stdio.h>
#pragma hdrstop

#include "Viewer.h"
#include "decode.h"
#include "Led.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"

TTestForm *TestForm;
/*	Formats:
1.  Normal
    Byte 0 - status = 0;
    Byte 1 - pupil diameter Most Significant Byte (MSB)
    BYTE 2 - pupil diameter Least Significant Byte (LSB)
    Byte 3 - Scene plane number.
    Byte 4 - Horizontal pos. MSB
    Byte 5 - Horizontal pos LSB
    Byte 6 - Vertical pos. MSB
    Byte 7 - Vertical pos. LSB
2. Target Points:
    Byte 0 - status = 2;
    Byte 1 - Target Point number (MSB)
    BYTE 2 - Target Point number (LSB)
    Byte 3 - Horizontal coord. MSB
    Byte 4 - Horizontal coord LSB
    Byte 5 - Vertical coord. MSB
    Byte 6 - Vertical coord. MSB
3. Uncalibrated:
    Byte 0 - status = 3;
    Byte 1 - pupil diameter Most Significant Byte (MSB)
    BYTE 2 - pupil diameter Least Significant Byte (LSB)
    Byte 3 - Scene plane number.
    Byte 4 - Horizontal pos. MSB
    Byte 5 - Horizontal pos LSB
    Byte 6 - Vertical pos. MSB
    Byte 7 - Vertical pos. LSB
    Byte 8 - Raw pupil horiz. pos. MSB
    Byte 9  - Raw pupil horiz. pos. LSB
    Byte 10 - Raw pupil vert. pos. MSB
    Byte 11  - Raw pupil vert. pos. LSB
    Byte 12 - Raw cr horiz. pos. MSB
    Byte 13  - Raw cr horiz. pos. LSB
    Byte 14 - Raw cr vert. pos. MSB
    Byte 15  - Raw cr vert. pos. LSB
    Byte 16 - XDat MSB
    Byte 17 - XDat LSB
*/
#define DATA_READY_MESSAGE			WM_USER + 1000
#define BUFFER_OVERFLOWN_MESSAGE	WM_USER + 1001


const size_t TTestForm::MaxInfoBytes = MAX_BUFFER_SIZE - 1;
const unsigned char StreamingModeStatusMask = 0x7F; //  1011 1111

//  Status byte values
const unsigned char TTestForm::NormalMode = 0;
const unsigned char TTestForm::TargetPointsMode = 1;
const unsigned char TTestForm::UncalibratedMode = 2;

//  Packet lengths in diff. modes:
const unsigned char TTestForm::NormalModeLen = 8;
const unsigned char TTestForm::TargetPointsModeLen = 7;
const unsigned char TTestForm::UncalibratedModeLen = 18;

static const char *DataBufferMutexName = "DATABUFFER_MUTEX";
static const DWORD WaitTimeout = 200;	//	200 ms

//	Test thread: generates data buffers:
__fastcall TTestThread :: TTestThread():
   TThread(true)
{
	COM = new TCOM(Application);
	COM->initialize( "COM1", 57600);
}

void __fastcall TTestThread::Execute()
{
	int type = 130;
	int nRawBytes;
    int nOutBytes;
	short int num = 0;
	unsigned long written;
	while (!Terminated)
   {
		nRawBytes = build_serial_data_out_buffer(raw_buffer, type, num);
		encode_buffer(raw_buffer, out_buffer, nRawBytes );
        nOutBytes = nRawBytes + ( ( nRawBytes - 1 ) / 7 ) + 1;
		*out_buffer |= 0x80;
		COM->write_buffer(out_buffer, nOutBytes, &written);
      num++;
   }
}

//---------------------------------------------------------------------------
__fastcall TReceiveThr :: TReceiveThr( TTestForm* Owner, TCOM *com, HANDLE mutex):
	TThread( true ),
	guiWnd(Owner),
   	COM(com),
    bufferInUseMutex(mutex)
{
	buffers[0] = buffer1;
	buffers[1] = buffer2;
	bufferSelect = 0;
	currentBuffer = buffers[bufferSelect];
	curDataPtr = currentBuffer;
}

void __fastcall TReceiveThr::Execute()
{
	DWORD waitresult;
	do
    {
       	if( COM->get_input_count() >0 )
        {
            unsigned char byte = COM->read_byte();
            if( byte & 0x80)
            {
                //	get the real length of the buffer:
                size_t bufferLength = curDataPtr - currentBuffer + 1;
        		//	Check if the GUI thread is ready to accept a new buffer:
                if(WaitForSingleObject(bufferInUseMutex, 0) == WAIT_OBJECT_0)
				{
                    //	Post message and pass the buffer pointer to the GUI thread:
                    PostMessage(guiWnd->Handle, DATA_READY_MESSAGE, WPARAM(bufferLength ), LPARAM( currentBuffer) );
                    //	switch to the new buffer:
                    bufferSelect ^= 1;
					ReleaseMutex(bufferInUseMutex);
				}
				//	if the thread was not ready to accept the buffer, keep overwriting it:
                currentBuffer = buffers[bufferSelect];
                curDataPtr = currentBuffer;
            }

            //	put the byte into the buffer, if not overflowing buffer:
            if(curDataPtr - currentBuffer < MAX_BUFFER_SIZE)
                *curDataPtr++ = byte;
            else
            {
                PostMessage( guiWnd->Handle, BUFFER_OVERFLOWN_MESSAGE, 0, LPARAM( currentBuffer) );
               	curDataPtr = currentBuffer;

            }
		}
	}
	while( !Terminated);
}

//---------------------------------------------------------------------------
__fastcall TTestForm::TTestForm(TComponent* Owner): TForm(Owner),
    testThread(0)
{
#ifdef LOOPBACK_TEST
	testThread = new TTestThread;
#endif
	//	Create mutex for data buffers:
    mutexSecAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    mutexSecAttr.lpSecurityDescriptor = 0;
    mutexSecAttr.bInheritHandle = false;

	bufferInUseMutex = CreateMutex(&mutexSecAttr, true, DataBufferMutexName);
    ReleaseMutex(bufferInUseMutex);

    COM = new TCOM( Application );
    ReceiveThr = new TReceiveThr( this, COM, bufferInUseMutex);
    FIniFile   = new TIniFile( ".\\viewer.ini" );

    showData = false;
    showMode = Stop;

    BoxCOM->ItemIndex = BoxCOM->Items->IndexOf( FIniFile->ReadString( "PORT","COM","COM2") );
    BoxRate->ItemIndex = BoxRate->Items->IndexOf( FIniFile->ReadString( "PORT","BaudRate","57600") );

    size_t protocolIndex = ProtocolSelectRG->Items->IndexOf( FIniFile->ReadString("PROTOCOL","Type","On Demand") );

    protocolType = protocolIndex == 0 ? Streaming : OnDemand;
    ProtocolSelectRG->ItemIndex = protocolIndex;
}

//---------------------------------------------------------------------------
void __fastcall TTestForm::FormClose(TObject *Sender, TCloseAction &Action)
{
    FIniFile->WriteString( "PORT","COM",BoxCOM->Text);
    FIniFile->WriteString( "PORT","BaudRate", BoxRate->Text);
    FIniFile->WriteString( "PROTOCOL","Type", ProtocolSelectRG->Items->Strings[protocolType]);

    if ( ReceiveThr && !ReceiveThr->Suspended)
    {
        COM->close();
        ReceiveThr->Terminate();
    }
}

void __fastcall TTestForm::WndProc(Messages::TMessage &Message)
{
	//	if not in streaming mode, do the default:
	if(protocolType != Streaming || !showData)
	{
		TForm::WndProc(Message);
        return;
	}

	if(Message.Msg == DATA_READY_MESSAGE)
	{
		//	Get the buffer-in-use mutex:
       	WaitForSingleObject(bufferInUseMutex, WaitTimeout);
		unsigned char *buffer = (unsigned char *)Message.LParam;
		memcpy(rawDataBuffer, buffer, MAX_BUFFER_SIZE);

        //  mask out streaming mode bit
        *rawDataBuffer = *rawDataBuffer & StreamingModeStatusMask;
        //  define the length of the packet to decode:
        switch(*rawDataBuffer)
        {
            case 0://NormalMode:
            {
                dataLen = NormalModeLen;
                 break;
            }
            case 1://TargetPointsMode:
            {
                dataLen = TargetPointsModeLen;
                break;
            }
            case 2://UncalibratedMode:
            {
                dataLen = UncalibratedModeLen;
                break;
            }
            default:
            {
                dataLen = 0;
            }
        }
        if(dataLen)
		{
    		decode_buffer(rawDataBuffer, dataBuffer, dataLen);
			ShowDataBuffer();
        }

        //  block further data display if single
		if(showMode == Single)
    		showData = false;

		ReleaseMutex(bufferInUseMutex);
		return;
    }
	else if (Message.Msg == BUFFER_OVERFLOWN_MESSAGE)
	{
        //  not processed yet.
		return;
	}
    TForm::WndProc(Message);
}

void TTestForm::GetBufferOnDemand()
{
    //  send a byte, then receive a packet back:
    unsigned long actualRead;
    COM->reset();
    COM->write_byte(0);
    COM->read_buffer( dataBuffer, MAX_BUFFER_SIZE, &actualRead );
    switch(*dataBuffer)
    {
        case 0://NormalMode:
        {
            dataLen = NormalModeLen;
             break;
        }
        case 1://TargetPointsMode:
        {
            dataLen = TargetPointsModeLen;
            break;
        }
        case 2://UncalibratedMode:
        {
            dataLen = UncalibratedModeLen;
            break;
        }
        default:
        {
            dataLen = 0;
        }
    }
}

void TTestForm::ShowDataBuffer()
{
    ReportPacketLength(dataLen);
    switch(*dataBuffer)
    {
        case 0://NormalMode:
        {
            ReportStatus("Normal");
            if(dataLen != NormalModeLen)
            {
                ReportError("NormalMode: Bad Length");
                break;
            }
            PageControl1->ActivePage = Normal;
            PD=IntToStr( (dataBuffer[1]<<8) + dataBuffer[2] );
            SP=IntToStr(  dataBuffer[3] );
            X=IntToStr( (dataBuffer[4]<<8) + dataBuffer[5] );
            //Some filtering for X coordinate
            if(StrToInt(X)<0)
              X="0";
            if(StrToInt(X)>261)
              X="261";
            Y=IntToStr( (dataBuffer[6]<<8) + dataBuffer[7] );
            //Some filtering for Y coordinate
            if(StrToInt(X)<0)
              Y="0";
            if(StrToInt(X)>261)
              Y="241";
            current_X=StrToInt(X)-StrToInt(X_top_left_corner);
            //Some filtering on the frame position for X
            if(current_X<0)
              current_X=0;
            if(current_X>StrToInt(Frame_Width))
              current_X=StrToInt(Frame_Width);
            current_Y=StrToInt(Y)-StrToInt(Y_top_left_corner);
            //Some filtering on the frame position for Y
            if(current_Y<0)
              current_Y=0;
            if(current_Y>StrToInt(Frame_Height))
              current_Y=StrToInt(Frame_Height);
            //Some filtering for Frame Width and Height
            if(StrToInt(Frame_Width)==0)
              Frame_Width="1";
            if(StrToInt(Frame_Height)==0)
              Frame_Height="1";
            current_X=(int)(current_X*100.0/StrToInt(Frame_Width));
            current_Y=(int)(current_Y*100.0/StrToInt(Frame_Height));
            Label17->Caption=IntToStr(current_X);
            Label19->Caption=IntToStr(current_Y);
            temp_count++;
            if(temp_count==10)
            {
              TListItem* ListItem = ListView1->Items->Add();
              if(ListView1->Items->Count > 18)
                ListView1->Items->Delete(0);
              ListItem->Caption=PD;
              ListItem->SubItems->Add(SP);
              ListItem->SubItems->Add(X);
              ListItem->SubItems->Add(Y);
              temp_count=0;
            }
            //
            fprintf(outfile,"%s  %s  %s  %s\n",PD,SP,X,Y);
            //
            sprintf(buf,"%d  %d\n",current_X,current_Y);
            if(Player_Connection)
            {
              int err=send(sockfd,buf,sizeof(buf),0);
              PrintList(IntToStr(err));
            }
            break;
        }
        case 1://TargetPointsMode:
        {
            ReportStatus("Target Point");
            if(dataLen != TargetPointsModeLen)
            {
                ReportError("TargetPoints: Bad Length");
                break;
            }

            PageControl1->ActivePage = TargetPoints;
            if( ListView3->Items->Count > 100 )
              ListView3->Items->Clear();
            TListItem* ListItem = ListView3->Items->Insert(0);
            ListItem->Caption = IntToStr( (dataBuffer[1]<<8) + dataBuffer[2] );
            ListItem->SubItems->Add(IntToStr( (dataBuffer[3]<<8) + dataBuffer[4]  ));
            ListItem->SubItems->Add(IntToStr( (dataBuffer[5]<<8) + dataBuffer[6]  ));
            break;
        }

        case 2://UncalibratedMode:
        {
            ReportStatus("Uncalibrated");
            if(dataLen != UncalibratedModeLen)
            {
                ReportError("Uncalibrated: Bad Length");
                break;
            }

            PageControl1->ActivePage = Uncalibrated;
            int tmpInt;

            WORD pupilHor, pupilVert, crHor, crVert;	//	raw data

            char tmp[10];
            if( ListView2->Items->Count > 1000 )
              ListView2->Items->Clear();

            TListItem* ListItem = ListView2->Items->Insert(0);

            tmpInt = MAKEWORD(dataBuffer[2],dataBuffer[1]);
            ListItem->Caption=itoa(tmpInt,tmp,10 );
            ListItem->SubItems->Add(IntToStr( dataBuffer[3] ));
            tmpInt = MAKEWORD(dataBuffer[5],dataBuffer[4]);
            ListItem->SubItems->Add(itoa( tmpInt, tmp,10 ));
            tmpInt = MAKEWORD(dataBuffer[7],dataBuffer[6]);
            ListItem->SubItems->Add(itoa(tmpInt, tmp,10 ) );
            //	raw data:
            pupilHor = MAKEWORD(dataBuffer[9],dataBuffer[8]);
            pupilVert = MAKEWORD(dataBuffer[11],dataBuffer[10]);
            crHor = MAKEWORD(dataBuffer[13],dataBuffer[12]);
            crVert = MAKEWORD(dataBuffer[15],dataBuffer[14]);

            tmpInt = pupilHor - crHor;
            ListItem->SubItems->Add(itoa(tmpInt, tmp,10 ) );
            tmpInt = pupilVert - crVert;
            ListItem->SubItems->Add(itoa(tmpInt, tmp,10 ) );
            //	xdat:
            sprintf(tmp,"%X", MAKEWORD(dataBuffer[17],dataBuffer[16]));
            ListItem->SubItems->Add(tmp);
            break;
        }
        default:
        {
            ReportStatus("");
            ReportError("Unknown Data Type");
            break;
        }
    }
}

void TTestForm::ReportError(const AnsiString &str)
{
    StatusBar->Panels->Items[2]->Text = str;
}

void TTestForm::ReportStatus(const AnsiString &str)
{
    StatusBar->Panels->Items[0]->Text = str;
}

void TTestForm::ReportPacketLength(size_t length)
{
    StatusBar->Panels->Items[1]->Text = "Bytes: "+IntToStr(length);
}


//---------------------------------------------------------------------------
void __fastcall TTestForm::BoxCOMChange(TObject *Sender)
{
    COM->close();
//    Led1->LedOn = false;
    COM->COM = BoxCOM->Text;
}

//---------------------------------------------------------------------------
void __fastcall TTestForm::BoxRateChange(TObject *Sender)
{
    COM->BaudRate= StrToInt( BoxRate->Text );
}

//---------------------------------------------------------------------------
void __fastcall TTestForm::BtnConnClick(TObject *Sender)
{
    if( !COM->initialize( BoxCOM->Text, StrToInt( BoxRate->Text )))
    {
 //       Led1->LedOn = false;
        StatusBar->Panels->Items[0]->Text = "error of port initialization";
    }
    else
    {
        ReportStatus("COM initializited");
        ReportError("");
        ReportPacketLength(0);

        if( ReceiveThr->Suspended && protocolType == Streaming)
            ReceiveThr->Resume();

//        Led1->LedOn = true;
    }
    if(testThread)
        testThread->Resume();

}

//---------------------------------------------------------------------------

void __fastcall TTestForm::Led1Change(TObject *Sender, bool IsOn)
{
//    Led1->Hint= Led1->LedOn?"COM connected":"COM not connected";
}


void __fastcall TTestForm::BtnSingleClick(TObject *Sender)
{
    showMode = Single;
    showData = true;
    if(protocolType == OnDemand)
    {
        GetBufferOnDemand();
        ShowDataBuffer();
    }
}

//---------------------------------------------------------------------------
void __fastcall TTestForm::BtnStartClick(TObject *Sender)
{
    outfile = fopen("data.out","w");
    showMode = Continuous;
    showData = true;

    BtnStop->Enabled = true;
    BtnStart->Enabled = false;
    BtnSingle->Enabled = false;

    if(protocolType == OnDemand)
    {
        while(showData && protocolType == OnDemand)
        {
            GetBufferOnDemand();
            ShowDataBuffer();
            //	let messages to be processed:
            Application->ProcessMessages();
        }
    }

}

//---------------------------------------------------------------------------
void __fastcall TTestForm::BtnStopClick(TObject *Sender)
{
    showMode = Stop;
    showData = false;
    BtnStart->Enabled = true;
    BtnSingle->Enabled = true;
    //
    fclose(outfile);
    //
}

void __fastcall TTestForm::ClearBTNClick(TObject *Sender)
{
	if(PageControl1->ActivePage == Normal)
    	ListView1->Items->Clear();
	else if(PageControl1->ActivePage == Uncalibrated)
    	ListView2->Items->Clear();
	else if(PageControl1->ActivePage == TargetPoints)
    	ListView3->Items->Clear();
}

//---------------------------------------------------------------------------
void __fastcall TTestForm::OnProtocolSelect(TObject *Sender)
{
    ListView1->Items->Clear();
    ListView3->Items->Clear();
    ListView2->Items->Clear();

    protocolType = ProtocolSelectRG->ItemIndex == 0 ? Streaming : OnDemand;
    if(protocolType == Streaming)
        ReceiveThr->Resume();
    else
        ReceiveThr->Suspend();

}

void TTestForm::PrintList(const AnsiString &str)
{
    TListItem* ListItem = ListView4->Items->Add();
    ListItem->Caption=str;
    if(ListView4->Items->Count > 6)
       ListView4->Items->Delete(0);
}


void __fastcall TTestForm::Button1Click(TObject *Sender)
{
  Button3->Enabled = true;
  Button2->Enabled = true;
  Button1->Enabled = false;
  AnsiString SerIPaddr = Edit1->Text;
  AnsiString SerPort = Edit2->Text;
  char sIP[256];
  char sPort[256];
  sprintf(sIP,"%s",SerIPaddr);
  sprintf(sPort,"%s",SerPort);
  int port=atoi(sPort);
  PrintList("Connect");
    WORD wVersionRequested;
    WSADATA wsaData;
    int err;
    wVersionRequested = MAKEWORD( 2, 0 );
    err = WSAStartup( wVersionRequested, &wsaData );
    PrintList(IntToStr(err));

    if((sockfd = socket(PF_INET,SOCK_STREAM,0))<0)
    {
      PrintList("socket error - initialization falure");
      PrintList(IntToStr(sockfd));
      PrintList("Error number "+IntToStr(WSAGetLastError()));
      Button1->Enabled = true;
      Button2->Enabled = false;
      Button3->Enabled = false;
      Button4->Enabled = false;
    }
    else
    {
      PrintList("Socket Initialized");
      PrintList(IntToStr(sockfd));
    }
    memset(&serv,0,sizeof(serv));
    serv.sin_family = AF_INET;
    serv.sin_addr.s_addr=inet_addr(sIP);
    serv.sin_port=htons(port);
    if(connect(sockfd,(struct sockaddr *)&serv,sizeof(serv))<0)
    {
        PrintList("socket error - connection falure ");
        PrintList("Check server's IP address and Port number ");
        Button1->Enabled = true;        
        Button2->Enabled = false;
        Button3->Enabled = false;
        Button4->Enabled = false;
    }
    else
    {
        PrintList("Socket sucsessfully connected");
        PrintList("Connected to server " + SerIPaddr + "; Port number "+SerPort);
    }
/*
    strcpy(buf,"Hello");
    err=send(sockfd,buf,sizeof(buf),0);
    PrintList(IntToStr(err));
*/    


}
//---------------------------------------------------------------------------

void __fastcall TTestForm::Button2Click(TObject *Sender)
{
  Button4->Enabled = true;
  Button2->Enabled = false;
  PrintList("Start Transfer");
  Player_Connection=1;
}
//---------------------------------------------------------------------------

void __fastcall TTestForm::Button3Click(TObject *Sender)
{
  Button1->Enabled = true;
  Button3->Enabled = false;
  Button2->Enabled = false;
  Button4->Enabled = false;
  PrintList("Disconnect");
  Player_Connection=0;  
  closesocket(sockfd);
}
//---------------------------------------------------------------------------

void __fastcall TTestForm::Button4Click(TObject *Sender)
{
  Button2->Enabled = true;
  Button4->Enabled = false;
  PrintList("Stop Transfer");
  Player_Connection=0;  
}
//---------------------------------------------------------------------------


void __fastcall TTestForm::Button5Click(TObject *Sender)
{
  AnsiString a,b;
  a=X_top_left_corner=X;
  b=Y_top_left_corner=Y;
  Label10->Caption=a+", "+b;
}
//---------------------------------------------------------------------------

void __fastcall TTestForm::Button6Click(TObject *Sender)
{
  AnsiString a,b;
  Label11->Caption=X+", "+Y;
  a=X_bottom_right_corner=X;
  b=Y_bottom_right_corner=Y;
  Label11->Caption=a+", "+b;
}
//---------------------------------------------------------------------------

void __fastcall TTestForm::Button7Click(TObject *Sender)
{

  Frame_Width=IntToStr(StrToInt(X_bottom_right_corner)-
  StrToInt(X_top_left_corner));
  Label14->Caption=Frame_Width;

  Frame_Height=IntToStr(StrToInt(Y_bottom_right_corner)-
  StrToInt(Y_top_left_corner));
  Label15->Caption=Frame_Height;

}

//---------------------------------------------------------------------------


